\CUBIC.XPL	17-Jan-2005		Lenny Boreal	loren_blaney@idcomm.com
\Tumbling cube demo to show off 32-bit XPL0
\This uses VESA graphics (640x480 in 256 colors) and megabytes of memory.
\A 500 MHz (or faster) processor is recommended.
\Compile with XPLPX (XPX.BAT)
\
\A right handed, 3-D coordinate system is used with the origin (O) at the
\ upper-left corner of the display screen. Positive rotations are shown.
\
\
\			            Z (into screen)
\			          /
\			      ----.
\			      /    \
\			    /       V     /
\			O +--------------|---- X
\			  |              |
\			  |              V
\			  |
\			  |  /
\			<---'
\			  |
\			  Y
\

include	c:\cxpl\codes;		\run-time library routines (intrinsics)

def	StarN = 3000;		\number of stars in starfield
def	Edge = 201;		\length of edge of cube (pixels)
def	IntSize = 4;		\number of bytes in an integer (2 or 4)
def	TVMode=$101, TVW=640, TVH=480;	\256-color graphics
def	PMWidth = 150,		\Permormance Monitor dimensions and position
	PMHeight = 61,		\ (pixels)
	PMX=TVW-PMWidth-1, PMY=TVH-PMHeight;

int	CpuReg,			\pointer to array holding processor's registers
	DSeg,			\data segment: where our data gets loaded (para)
	FreeMem,		\amount of free memory (beyond global arrays)
	PenX, PenY,		\pen coordinates (pixels) in Image for drawing
	StartTime,		\starting time, used for diagnostic purposes
	FrameTime;		\frame rate (microseconds)

int	CubeX, CubeY, CubeZ,	\cube's center coordinates (pixels)
	FrameCount,		\about 30 Hz, depending on hardware
	PMFill,			\Performance Monitor's fill index into PMGraph
	Time0;			\time at start of a time delay (838ns 16 bit)
real	CubeVx, CubeVy, CubeVz,	\cube's velocity (pixels per frame)
	CubeSize;		\cube's magnification factor (function of dist.)
real	CornerX, CornerY, CornerZ; \arrays: coordinates of cube's corners

char	PMGraph(PMWidth),	\Performance Monitor graph data; circular buffer
	Bkgnd(TVW*TVH),		\background image for blueprint
	Image(TVW*TVH),	  	\buffer for setting up copy of screen image
	Face(6, Edge*Edge); 	\pixel image (texture map) of each face
int	StarX(StarN), StarY(StarN), StarZ(StarN); \3D coordinates for stars

\Variables for Life:
char	LifeGen(40,40),		\generation displayed: cells are 5x5 pixels
	NextGen(40,40);		\ Edge/5 = 201/5 = 40

\Variables for Listing (also Scroller):
int	CurX, CurY,		\text cursor position
	FontHeight,		\character height in pixels = number of bytes
				\ per character (an 8x8 font is assumed)
	FontOff,		\address of BIOS font table (offset and segment)
	FontSeg;
char	NoiseImage(320*400);

\Variables for Plot3D waveform:
int	MovieFrame;		\movie frame counter
def	MovieFrames = 50;	\number of movie frames
char	Movie(MovieFrames, Edge*Edge);

\Variables for Kaleidoscope:
int	KaleidX, KaleidY,	\coordinates of pen within cube face
	KaleidColor;		\color of pen

\Variables for Scroller:
int	ScrollFrameCnt,		\frame count
	ScrollMsgX;		\horizontal coordinate of beginning of message
char	ScrollImage(Edge*Edge);	\array for setting up copy of screen image

\Variables for Molecule:
def	Atoms = 9;		\number of atoms (spheres)
int	AtomC(Atoms);		\color of each atom (0, 1, or 2)
real	AtomX(Atoms), AtomY(Atoms), AtomZ(Atoms), \coordinates of centers
	AtomR(Atoms);		\radii of atoms (atomic units)
int	Bond;			\array of indexes; indicates bonds between atoms
char	Sphere(81*81);		\image of a shaded sphere (color is added later)

def	Tab=$09, LF=$0A, CR=$0D, EOF=$1A;
def	Black, Blue, Green, Cyan, Red, Magenta, Brown, White,  \Attribute colors
	Gray, LBlue, LGreen, LCyan, LRed, LMagenta, Yellow, BWhite; \EGA palette

\-------------------------------------------------------------------------------

func	CallInt(Int, AX, BX, CX, DX, BP, DS, ES); \Call software interrupt
int	Int, AX, BX, CX, DX, BP, DS, ES; \(Unused arguments need not be passed)
begin
Cpureg(0):= AX;
Cpureg(1):= BX;
Cpureg(2):= CX;
Cpureg(3):= DX;
Cpureg(6):= BP;
Cpureg(9):= DS;
Cpureg(11):= ES;
Softint(Int);
return Cpureg(0);		\return AX register
end;	\CallInt



func	ShiftKey;		\Returns 'true' if shift key is down or CapsLock
return (CallInt($16, $0200) & $43) # 0;	\Function $02



proc	OpenTimer;		\Initialize the (8253) system timer
\The system timer (timer 0) is normally set to mode 3, as defined by IBM.
\ Some BIOSes, Win3.1 and WinXP have a bug that instead sets this timer to
\ mode 2. This does not change the interrupt rate, but it does change the
\ rate that the internal counter decrements. Mode 3 decrements by 2, and
\ mode 2 decrements by 1. Since the internal counter is read by this code,
\ it is essential that the rate be correct. Of course WinXP (in its infinite
\ wisdom) does not allow reprogramming the mode, so the non-standard (but
\ more logical) mode 2 is used here.
begin
port($43):= $34;		\set timer 0 to mode 2
port($40):= $FF;		\set 16-bit countdown timer to maximum count
port($40):= $FF;		\ (low byte first)
end;	\OpenTimer;



proc	CloseTimer;		\Restore normal timer mode (3)
begin
port($43):= $36;
port($40):= $FF;		\set to maximum count
port($40):= $FF;		\ (low byte first)
end;	\CloseTimer



func	GetTimeX;		\Get time from the 8253 counter
int	T;
begin				\each count is 838 ns (= 1 / 1.19e6 Hz)
port($43):= 0;			\latch counter 0
T:= port($40) + port($40)<<8;	\read 2 bytes
return T | $0000FFFF;		\reinvert the bits
end;	\GetTimeX



proc	Delay(TD);		\Time delay
int	TD;	\time in microseconds (maximum is 26000)
int	Cnt, T0;
begin
Cnt:= TD*1000/838;
T0:= GetTimeX;
repeat until ((GetTimeX-T0) & $FFFF) >= Cnt;  \25ms = 25e6ns; 25e6ns/838 = 29833
end;	\Delay



proc	Wait26ms;		\Wait 26 milliseconds since Time0 was read
\26ms = 26e6ns; 26e6ns/838 = 31026
repeat until ((GetTimeX-Time0) & $FFFF) >= 31026;



func	ElapsedTime;		\Returns time elapsed in microseconds since
int	T;			\ StartTime:= GetTimeX
begin				\ (must be <= 54,919 microseconds)
T:= GetTimeX;
T:= T - StartTime;
if T < 0 then T:= T + 65536;
return T*838/1000;
end;	\ElapsedTime

\-------------------------------------------------------------------------------

proc	Exit(Msg);		\Display (error) message and terminate program
char	Msg;
begin
SetVid(3);					\restore normal text mode
Text(0, Msg);
Crlf(0);
Text(0, "Hit a key to continue...");		\necessary for Windoze
Msg:= Chin(1);
CloseTimer;
exit;
end;	\Exit



proc	WaitVB;			\Wait for vertical blank
begin
while port($3DA) & $08 do;	\wait for vertical blank to go away
repeat until port($3DA) & $08;	\wait for vertical blank
end;	\WaitVB



proc	Fadeout;		\Fade all colors to black
int	I, J, C;
char	Pal(256*3);
begin
port($3C7):= 0;					\read color registers
for J:= 0, 256*3-1 do Pal(J):= port($3C9);

for I:= 0, 63-1 do				\decrement intensities 63 times
	begin
	for J:= 0, 256*3-1 do
		begin
		C:= Pal(J) - 1;
		if C < 0 then C:= 0;
		Pal(J):= C;
		end;
	WaitVB;
	port($3C8):= 0;				\write color registers
	for J:= 0, 256*3-1 do port($3C9):= Pal(J);
	end;
end;	\Fadeout

\############################ PERFORMANCE MONITOR ##############################

proc	DrawChar(Ch, C);	\Draw a character at PenX,PenY on Image
int	Ch, C;			\upper-left coordinates, character
int	I, J, Byte;		\coordinates, size, indicies, font byte
begin
for J:= 0, FontHeight-1 do
	begin			\get byte from BIOS's font table
	Byte:= Peek(FontSeg,  FontOff + Ch*FontHeight + J);
	for I:= 0, 7 do		\for each bit in the Byte...
		begin
		if Byte & $80 then
			Image(PenX+I + (PenY+J)*TVW):= C;
		Byte:= Byte << 1;
		end;
	end;
PenX:= PenX + 8;		\move to next character position
end;	\DrawChar



proc	DrawText(Str, C);	\Draw all the characters in a text string
char	Str;
int	C;
int	I, Ch;
begin
for I:= 0, 1000 do
	begin
	Ch:= Str(I);
	if Ch >=$80 then
		[Ch:= Ch & $7F;   I:= 1000];
	DrawChar(Ch, C);
	end;
end;	\DrawText



proc	DrawNum(N, C);		\Draw a positive integer at PenX,PenY on Image
int	N, C;
int	Q, R;
begin
Q:= N/10;
R:= rem(0);
if Q \#0\ then DrawNum(Q, C);
DrawChar(R+^0, C);
end;	\DrawNum



proc	DrawLine(X, Y, C);	\Draw line on Image from PenX,PenY to X,Y
int	X, Y, C;	\coordinates of end-of-line, color
int	DX, DY,		\delta X, delta Y
	IX, IY,		\plotting increments along X and Y axes
	SY,		\flag: step along Y direction
	D,		\decision variable (move straight or diagonally)
	E, NE,		\East, Northeast (straight or diagonal)
	T, I;		\scratch
begin
DX:= X - PenX;				\delta X
if DX >= 0 then				\if positive then increment
	IX:= 1
else	[DX:= -DX;   IX:= -1];		\else absolute value and decrement

DY:= Y - PenY;				\same for Y direction
if DY >= 0 then
	IY:= 1
else	[DY:= -DY;   IY:= -1];

if DX >= DY then			\step along X direction
	SY:= false
else	begin
	SY:= true;			\step along Y direction
	T:= DX;   DX:= DY;   DY:= T;	\swap DX and DY
	end;

D:= 2*DY - DX;				\set up decision variable stuff
E:= 2*DY;
NE:= 2*(DY-DX);

Image(PenX + PenY*TVW):= C;		\plot initial point
for I:= 1, DX do
	begin
	if D < 0 then
		begin			\move straight
		D:= D + E;
		if SY then PenY:= PenY + IY
		      else PenX:= PenX + IX;
		end
	else	begin			\move diagonally
		D:= D + NE;
		PenX:= PenX + IX;
		PenY:= PenY + IY;
		end;
	Image(PenX + PenY*TVW):= C;
	end;
end;	\DrawLine



proc	DrawRect(X0, Y0, X1, Y1, C);	\Draw a filled rectangle into Image
int	X0, Y0, X1, Y1, C;
int	Y;
begin
for Y:= Y0, Y1 do
	begin
	PenX:= X0;   PenY:= Y;
	DrawLine(X1, Y, C);
	end;
end;	\DrawRect



proc	DrawPM(T);	\Draw the Performance Monitor onto Image
int	T;		\time, converted from us to ms and plotted on graph
int	I, X, Y;
begin
Y:= (T+500)/1000;	\convert to nearest millisecond
if Y > PMHeight-1 then Y:= PMHeight-1;	\limit maximum value
PMGraph(PMFill):= Y;	\store new value in PMGraph
PMFill:= PMFill + 1;
if PMFill >= PMWidth then PMFill:= 0;

DrawRect(PMX, PMY, PMX+PMWidth-1, PMY+PMHeight-1, $12);	\dark gray
for Y:= 0, PMHeight/10 do
	begin
	PenX:= PMX;   PenY:= PMY+PMHeight-10*Y-1;
	DrawLine(PMX+PMWidth-1, PMY+PMHeight-1-10*Y, $14);
	end;

PenX:= PMX-1;   PenY:= PMY;
DrawLine(PMX-1, PMY+PMHeight-1, $14);
PenX:= PMX+PMWidth;   PenY:= PMY;
DrawLine(PMX+PMWidth, PMY+PMHeight-1, $14);

\Draw red line at desired frame rate (33 ms for a 60Hz display)
if FrameTime > 0 then
	begin
	Y:= 0;
	repeat Y:= Y+FrameTime until Y >= 26000;	\see Wait26ms
	Y:= (Y+500)/1000;	\round to nearest millisecond
	PenX:= PMX;   PenY:= PMY+PMHeight-1-Y;
	DrawLine(PMX+PMWidth-1, PMY+PMHeight-1-Y, Red);
	end;

I:= PMFill;
for X:= 0, PMWidth-1 do
	begin
	Y:= PMGraph(I);
	I:= I + 1;
	if I >= PMWidth then I:= 0;
	Image(X+PMX + (PMY+PMHeight-1-Y)*TVW):= $B;
	end;

PenX:= PMX+2; PenY:= PMY+2;		\display time
DrawNum(T, Yellow);
DrawChar($E6, Yellow); DrawChar(^s, Yellow);

if FrameTime > 0 then
	begin		\display video hardware frames per second (rounded)
	PenX:= PMX+2; PenY:= PMY+PMHeight-1-FontHeight-10;
	DrawNum((1000000+FrameTime/2)/FrameTime, Yellow);
	DrawText("Hz", Yellow);
	end;

PenX:= PMX+2; PenY:= PMY+PMHeight-1-FontHeight;
DrawNum(FreeMem/(1024*1024), Yellow);
DrawText("MB", Yellow);
end;	\DrawPM



proc	OpenPM;
begin
for PMFill:= 0, PMWidth-1 do
	PMGraph(PMFill):= 0;
PMFill:= 0;
end;

\################################## LIFE #######################################

proc	DrawCell(X, Y, Cell);	\Draw a Life cell in a 4x4 pixel area
int	X, Y,	\LifeGen array coordinates
	Cell;	\0=dead; 1=alive


	proc	DrawPix(X, Y, C, DX, DY);
	int	X, Y,	\screen coordinates (pixels)
		C,	\color
		DX, DY;	\deltas
	begin
	Face(1, (Y+DY)*Edge+X+DX):= C;
	Face(1, (Y+DY)*Edge+X-DX-1):= C;
	Face(1, (Y-DY-1)*Edge+X+DX):= C;
	Face(1, (Y-DY-1)*Edge+X-DX-1):= C;
	end;	\DrawPix


begin				\shades of gray
if Cell # 0 then
	begin
	DrawPix(X*5+1+2, Y*5+1+2, $10, 0, 0);
	DrawPix(X*5+1+2, Y*5+1+2, $1A, 1, 1);
	DrawPix(X*5+1+2, Y*5+1+2, $13, 1, 0);
	DrawPix(X*5+1+2, Y*5+1+2, $13, 0, 1);
	end;
end;	\DrawCell



proc	CycleLife;
\Display one generation of John Conway's cellular automaton.
\Rules:
\ A cell is born if it is adjacent to exactly three live cells.
\ A cell dies if it is adjacent to more than three or less than two cells.
\ "Adjacent" means the eight surrounding cells.
int	FA, X, Y, I, J, N;
begin
FA:= addr Face(1, 0);			\fill Face's background
for I:= 0, Edge*Edge/IntSize-1 do	\write four bytes at a time for speed
	FA(I):= $1D1D1D1D;		\fill with background color
Face(1, Edge*Edge-1):= $1D;		\erase last pixel (in case Edge is odd)

for Y:= 0, 40-1 do			\display LifeGen array
	for X:= 0, 40-1 do
		DrawCell(X, Y, LifeGen(X,Y));

for Y:= 1, 40-2 do			\determine next generation
	begin
	for X:= 1, 40-2 do
		begin
		N:= 0;			\count surrounding cells
		for J:= Y-1, Y+1 do
		    for I:= X-1, X+1 do
			begin
			if I#X ! J#Y then
				if LifeGen(I, J) \#0\ then N:= N + 1;
			end;
		if N = 3 then NextGen(X, Y):= 1			\birth
		else if N>=4 ! N<=1 then NextGen(X, Y):= 0	\death
		else NextGen(X,Y):= LifeGen(X,Y);
		end;
	end;
I:= LifeGen;   LifeGen:= NextGen;   NextGen:= I;	\swap LifeGen & NextGen
end;	\CycleLife



proc	InitLife;		\Initialize Life
int	X, Y;
char	Data;
begin
for Y:= 0, 40-1 do		\zero arrays
    for X:= 0, 40-1 do
	begin
	LifeGen(X, Y):= 0;
	NextGen(X, Y):= 0;
	end;

\Set up starting patterns (glider gun and R pentomino):
Data:= ["........................................",
	".......................##...............",
	".........................#..............",
	"............#.............#.............",
	"............##............#........#....",
	".........##...............#........##...",
	".#......###..............#..............",
	".##......##............##...............",
	"............##..........................",
	"............#..........................."];

for Y:= 0, 10-1 do
    for X:= 0, 40-1 do
	LifeGen(X, Y):= if Data(Y,X) = ^# then 1 else 0;

Data:= ["............##..........................",
	"...........##...........................",
	"............#..........................."];

for Y:= 30, 33-1 do
    for X:= 0, 40-1 do
	LifeGen(X, Y):= if Data(Y-30,X) = ^# then 1 else 0;
end;	\InitLife

\################################## NOISE ######################################

proc	Noise(F);		\Image full of white noise
int	F;	\face to display on
int	XOff, YOff, X, Y;
begin
XOff:= Ran(320);		\pick a random offset into array of dots
YOff:= Ran(200);
Blit(DSeg, NoiseImage+XOff+YOff*320, DSeg, addr Face(F,0), Edge*Edge);
end;	\Noise



proc	InitNoise;		\Fill array with random dots in shades of gray
int	I;
begin
for I:= 0, 320*400-1 do
	NoiseImage(I):= Ran(16)+16;
end;	\InitNoise

\################################# LISTING #####################################

proc	Listing;  \List the source code in a 40x25-character area on cube face



proc	Scroll;			\Scroll text up one line
int	X, Y;
begin
\Move 24 lines of text up one text line. Each text line is 8 pixels high.
Blit(DSeg, addr Face(3,0)+8*Edge, DSeg, addr Face(3,0), 24*8*Edge);

for Y:= 24*8, 25*8-1 do		\blank bottom line of text
    for X:= 0, Edge-1 do
	Face(3, X+Y*Edge):= $12;
end;	\Scroll



proc	DispCh(Ch);		\Display character using 8x8 font shrunk to 5x8
\i.e: use every other dot in BIOS's 8x8 font plus space between characters
int	Ch, I, J, Byte;
begin
for J:= 0, FontHeight-1 do
	begin			\get byte from BIOS's font table
	Byte:= Peek(FontSeg,  FontOff + Ch*FontHeight + J);
	for I:= 0, 3 do		\for each pair of bits in the font Byte...
		begin
		case Byte & $C0 of
		  $C0:		Face(3, CurX+I+(CurY+J)*Edge):= LGreen;
		  $80, $40:	Face(3, CurX+I+(CurY+J)*Edge):= Green
		other		[];
		Byte:= Byte << 2;
		end;
	end;
end;	\DispCh



proc	OutCh(Ch);		\Output a character to the display
int	Ch;
begin
case Ch of
  CR:	CurX:= 0;
  LF:	Scroll
other	if CurX < 40*5 then	\(5*40 = 200, which approximately equals Edge)
		[DispCh(Ch);   CurX:= CurX + 5];
end;	\OutCh



func	GetCh;			\Get character from input file
int	Ch;
begin
Ch:= Chin(3);
if Ch=Tab then Ch:= $20;	\convert Tab to space (width is only 40 chars)
if Ch = EOF then		\start over
	[OpenI(3);   Ch:= GetCh];
return Ch;
end;	\GetCh



begin	\Listing
OutCh(GetCh);
end;	\Listing



proc	InitListing;		\Initialize Listing
int	Hdl;	\input file handle
begin
\Open the source code file
Trap(false);			\don't trap "file not found" error; we'll do it
Hdl:= FOpen("CUBIC.XPL", 0);	\open file for input
if Geterr \#0\ then Exit("CUBIC.XPL not found");
Trap(true);
FSet(Hdl, ^I);			\set device 3 to handle
OpenI(3);
end;	\InitListing

\################################## PLOT3D #####################################

proc	Plot3D;			\Play movie of plotted 3D waveforms
int	IF, IM;
begin
IF:= Face;		\put addresses of char arrays into integers so we can
IM:= Movie;		\ copy the address of the frame to display (there is no
IF(2):= IM(MovieFrame);	\ need to copy the individual pixels)
MovieFrame:= MovieFrame + 1;
if MovieFrame >= MovieFrames then MovieFrame:= 0;
end;	\Plot3D



proc	InitPlot3D;		\Cheat a little and do all the calculations here
int	II;			\ and make the resulting displays into a movie
real	Time;			\time corresponding to frame image
def	Pi2 = 3.14159265358979323846 * 2.0;



proc	Plot3D(N);
\Translated from a BASIC program by Mark Bramhall of DEC.
int	N;			\movie frame number
def	Cx=100, Cy=120;		\offsets to center image
def	R=45.0, Mag=2.1;	\radius of base circle and image magnification
int	I, J;			\coordinates to plot
real	DX, DY,			\deltas
	X, Y, Y1, Z, Z1,	\coordinates
	D;			\distance from center
begin
for J:= 0, Edge-1 do		\clear movie frame
    for I:= 0, Edge-1 do
	Movie(N, I+J*Edge):= $12; \dark gray

DX:= 0.05;			\(small steps fill in holes in steep slopes)
DY:= 5.0;
X:= -R;
while X <= R do				\step from left to right
	begin
	Z1:= -1e10;
	Y1:= sqrt(R*R - X*X);		\base (in X-Y plane) is a circle
	Y:= Y1;
	while Y >= -Y1 do		\step from front to back
		begin
		D:= sqrt(X*X + Y*Y);
		Z:= 50.0*Cos(D/3.9-Time)*Exp(-D/15.0);
		Z:= Z - 0.707*Y;	\tilt forward 45 degrees
		if Z > Z1 then		\taller than previous value in front
			begin
			Z1:= Z;		\update tallest
			I:= fix(X*Mag) + Cx;
			J:= Cy - fix(Z*Mag);
			if I>=0 & I<Edge & J>=0 & J<Edge then	\clip
				Movie(N, I+J*Edge):= Blue;	\plot(I,J)
			end;
		Y:= Y - DY;
		end;
	X:= X + DX;
	end;
end;	\Plot3D



begin	\InitPlot3D
Time:= 0.0;			\start recording the movie
for II:= 0, MovieFrames-1 do
	begin
	Plot3D(II);
	Time:= Time + Pi2/float(MovieFrames);
	end;
MovieFrame:= 0;			\initialize for playback
end;	\InitPlot3D

\############################## KALEIDOSCOPE ###################################

proc	Kaleid;			\Kaleidoscope by Boyle, Wall, and Wasinger
\This draws rather nice kaleidoscopic patterns



	proc	Plot(X, Y, Color);	\Plot using eight-fold symmetry
	int	X, Y, Color;
	def	W=Edge/2, H=Edge/2;

		proc	Point(X, Y, C);	\Redefine Point intrinsic
		int	X, Y, C;
		Face(4, X+Edge*Y):= C;

	begin
	Point(X+W, Y+H, Color);
	Point(X+W, -Y+H, Color);
	Point(-X+W, -Y+H, Color);
	Point(-X+W, Y+H, Color);
	Point(Y+W, X+H, Color);
	Point(Y+W, -X+H, Color);
	Point(-Y+W, -X+H, Color);
	Point(-Y+W, X+H, Color);
	end;	\Plot


begin	\Kaleid
if Ran(40) = 0 then KaleidColor:= KaleidColor + 1;
if KaleidColor >= 20 then KaleidColor:= 0;	\use some dark colors beyond 15

KaleidX:= KaleidX + Ran(5) - 2;			\ +/-2, +/-1, or 0
KaleidY:= KaleidY + Ran(5) - 2;

KaleidX:= rem(KaleidX/100);			\keep point on screen (wraps)
KaleidY:= rem(KaleidY/100);

Plot(KaleidX, KaleidY, KaleidColor);		\eight-fold symmetry
end;	\Kaleid



proc	InitKaleid;
int	X, Y;
begin
KaleidX:= 0;			\start pen in the center
KaleidY:= 0;
KaleidColor:= 1;		\start with blue color

for Y:= 0, Edge-1 do		\blank the display area
    for X:= 0, Edge-1 do
	Face(4, X+Y*Edge):= $13; \dark gray
end;	\InitKaleid

\################################ SCROLLER #####################################

proc	Scroller;	\Yer basic scroller (with jumpin' background)
def	MsgLen = 30;	\number of characters in scroll message
int	XX, YY;		\coordinates (pixels)



proc	FillBkgnd;	\Fill background with moving geometric pattern
int	X, Y, Z;	\image coordinates (pixels), and index into ScrollImage
begin
Z:= 0;
for Y:= ScrollFrameCnt, ScrollFrameCnt+Edge-1 do			
    for X:= ScrollFrameCnt, ScrollFrameCnt+Edge-1 do
	begin
	ScrollImage(Z):= X ! Y | ScrollFrameCnt;	\jumpin' Sierpinskis!
	Z:= Z + 1;
	end;
end;	\FillBkgnd



proc	DrawTile(X0, Y0, S, C);	\Draw a square tile
int	X0, Y0, S, C;		\location (pixels), length of side, color
int	X, Y, J;
begin
if X0 >= Edge-S then return;	\clip to screen (X0 is always positive)

Y:= Y0 * Edge;
for J:= 0, S-1 do
	begin
	for X:= X0, X0+S-1 do ScrollImage(Y+X):= C;
	Y:= Y + Edge;
	end;
end;	\DrawTile



func	DispCh(X0, Y0, Ch);	\Display a character
int	X0, Y0, Ch;		\upper-left coordinates, character
int	X, Y, S, I, J, Byte;	\coordinates, size, indicies, font byte
begin
for J:= 0, FontHeight-1 do
	begin			\get byte from BIOS's font table
	Byte:= Peek(FontSeg,  FontOff + Ch*FontHeight + J);
	X:= X0;
	for I:= 0, 7 do		\for each bit in the Byte...
		begin
		if X < 0 then			\reflect small character
			begin
			S:= 2;			\size = small tiles
			Y:= J + Y0 + 75;	\start at base line
			end
		else	begin
			S:= 9;			\size = large tiles
			Y:= S*J + Y0;		\start at top
			end;
		if Byte & $80 then
			begin			\abs gives mirror image of char
			DrawTile(abs(X), Y, S, Black);	\outline
			DrawTile(abs(X+1), Y+1, S-2, BWhite);
			end;
		Byte:= Byte << 1;
		X:= X + S;			\move to next tile position
		end;
	end;
return X;			\return horizontal location for next char
end;	\DispCh



proc	DispMsg(X, Y, Msg);	\Display a message
int	X, Y;	\coordinates of start of message
char	Msg;	\pointer to message
int	I;
begin
for I:= 0, MsgLen-1 do
	X:= DispCh(X, Y, Msg(I));
end;	\DispMsg



begin	\Scroller
FillBkgnd;

DispMsg(ScrollMsgX, 70, "   ...Cubic in 32-bit XPL0!  ^B ");

ScrollMsgX:= ScrollMsgX - 1;		\shift large message to the left
ScrollFrameCnt:= ScrollFrameCnt - 1;	\make background pattern move

if ScrollMsgX < -(MsgLen*8*2 + Edge) then \small message is shifted off screen
	ScrollMsgX:= 0;			\start message over again

for YY:= 0, Edge-1 do			\copy ScrollImage onto cube Face
    for XX:= 0, Edge-1 do
	Face(0, XX+Edge*YY):= ScrollImage(XX+YY*Edge);
end;	\Scroller

\-------------------------------------------------------------------------------

proc	InitScroller;		\Initialize Scroller
begin
CpuReg(0):= $1130;	\get font table location
CpuReg(1):= $0300;	\8x8 font
SoftInt($10);
FontOff:= CpuReg(6) & $FFFF;
FontSeg:= CpuReg(11) & $FFFF;
FontHeight:= 8;

ScrollFrameCnt:= 36;			\initialize counter for moving backgnd
ScrollMsgX:= 0;				\initialize horizontal location of msg
end;	\InitScroller

\################################ MOLECULE #####################################

proc	RotateMole;		\Rotate each atom in the molecule
int	I;
def	Cx= 0.025, Cy= 0.03, Cz=0.013;		\determine rotation speed
begin
for I:= 0, Atoms-1 do				\rotate all atoms...
	begin
	AtomY(I):= AtomY(I) - AtomZ(I)*Cx;	\rotate about X axis
	AtomZ(I):= AtomZ(I) + AtomY(I)*Cx;
	AtomZ(I):= AtomZ(I) - AtomX(I)*Cy;	\rotate about Y axis
	AtomX(I):= AtomX(I) + AtomZ(I)*Cy;
	AtomX(I):= AtomX(I) - AtomY(I)*Cz;	\rotate about Z axis
	AtomY(I):= AtomY(I) + AtomX(I)*Cz;
	end;
end;	\RotateMole



proc	DrawSphere(X0, Y0, C0, R);	\Draw a shaded sphere onto cube Face
int	X0, Y0,	\coordinates of center (pixels)
	C0,	\1=yellow; 2=red; 3=cyan
	R;	\radius (pixels)
int	I, J, K, X, Y, C, CB;
begin
C0:= C0*64;			\64 shades for each color
for Y:= -R, R do
	begin
	J:= 40 + 40*Y/R;	\sphere pattern has radius 40
	for X:= -R, R do
		begin
		I:= 40 + 40*X/R;
		C:= Sphere(I+J*81);	\get color's intensity
		if C\#0\ then		\if on sphere then...
			begin
			K:= (Y+Y0)*Edge + X+X0;
			CB:= Face(5, K);		\get color of background
			CB:= ($3F - (CB&$3F)) >>2;	\remove hue; adj intense
			if C # $3F then C:= C - CB;	\don't alter highlight
			Face(5, K):= C+C0;		\plot color intense +hue
			end;
		end;
	end;
end;	\DrawSphere

\-------------------------------------------------------------------------------

proc	DrawMole;		\Draw one frame of the tumbling molecule
\Draw spheres, starting with the most distant ones. (Don't use perspective
\ because it causes an annoying jump in sphere sizes.)
def	Scale = 42.0;	\adjusts spacing between atoms
def	Cx = Edge/2-20,	\offset to center of molecule
	Cy = Edge/2-20;
int	I, J, L, B, X, Y, R, FarI,
	PX, PY,		\pen coordinates (pixels)
	Bx, By, Bc;	\offsets and colors for bond lines
real	Far;
int	Drawn(Atoms);



proc	LineTo(X, Y, C);	\Draw line from PX,PY to X,Y in color C
int	X, Y, C;	\coordinates of end-of-line, color
int	DX, DY,		\delta X, delta Y
	IX, IY,		\plotting increments along X and Y axes
	SY,		\flag: step along Y direction
	D,		\decision variable (move straight or diagonally)
	E, NE,		\East, Northeast (straight or diagonal)
	T, I;		\scratch
begin
DX:= X - PX;				\delta X
if DX >= 0 then				\if positive then increment
	IX:= 1
else	[DX:= -DX;   IX:= -1];		\else absolute value and decrement

DY:= Y - PY;				\same for Y direction
if DY >= 0 then
	IY:= 1
else	[DY:= -DY;   IY:= -1];

if DX >= DY then			\step along X direction
	SY:= false
else	begin
	SY:= true;			\step along Y direction
	T:= DX;   DX:= DY;   DY:= T;	\swap DX and DY
	end;

D:= 2*DY - DX;				\set up decision variable stuff
E:= 2*DY;
NE:= 2*(DY-DX);

Face(5, PX + PY*Edge):= C;		\plot initial point
for I:= 1, DX do
	begin
	if D < 0 then
		begin			\move straight
		D:= D + E;
		if SY then PY:= PY + IY
		      else PX:= PX + IX;
		end
	else	begin			\move diagonally
		D:= D + NE;
		PX:= PX + IX;
		PY:= PY + IY;
		end;
	Face(5, PX + PY*Edge):= C;
	end;
end;	\LineTo



proc	Erase(Im);	\Erase Face image buffer
int	Im;
int	I;
begin
for I:= 0, Edge*Edge/IntSize-1 do  \write four bytes at a time for speed
	Im(I):= $13131313;	\fill with dark gray
Face(5, Edge*Edge-1):= $13;	\erase last pixel (in case Edge is odd)
end;	\Erase



begin	\DrawMole
Erase(addr Face(5, 0));

\Draw bonds:
Bx:= [  2,  -2,   0,   0,   1,  -1,   0,   0,   0];
By:= [  0,   0,   2,  -2,   0,   0,   1,  -1,   0];
Bc:= [$18, $18, $18, $18, $1A, $1A, $1A, $1A, $1C];

for B:= 0, 16-1 do			\for bond end point...
	begin
	for L:= 0, 9-1 do		\bonds are drawn as 9 line segments
		begin
		I:= Bond(B);
		PX:= fix(AtomX(I)*Scale) + Cx + Bx(L);
		PY:= fix(AtomY(I)*Scale) + Cy + By(L);
		I:= Bond(B+1);
		X:= fix(AtomX(I)*Scale) + Cx + Bx(L);
		Y:= fix(AtomY(I)*Scale) + Cy + By(L);
		LineTo(X, Y, Bc(L));
		end;
	B:= B + 1;			\bond end points are in pairs
	end;

for I:= 0, Atoms-1 do			\mark all atoms as "not drawn"
	Drawn(I):= false;

for J:= 0, Atoms-1 do
	begin
	\Find the farthest atom that has not been drawn
	Far:= -1e12;
	for I:= 0, Atoms-1 do
		begin
		if AtomZ(I)>Far & not Drawn(I) then
			begin
			Far:= AtomZ(I);
			FarI:= I;
			end;
		end;

	\Draw the farthest atom that has not already been drawn
	I:= FarI;
	X:= fix(AtomX(I)*Scale) + Edge/2 - 20;
	Y:= fix(AtomY(I)*Scale) + Edge/2 - 20;
	R:= fix(AtomR(I) * 5.1);	\adjusts size of atoms
	DrawSphere(X, Y, AtomC(I), R);
	Drawn(I):= true;		\mark atom as "drawn"
	end;
end;	\DrawMole

\-------------------------------------------------------------------------------

proc	Molecule;		\Display a (colorful) tumbling molecule
begin
DrawMole;
RotateMole;
end;	\Molecule



proc	InitMole;		\Initialize data for Molecule
int	I, J, X, Y, Z, Z2, C;
def	R2 = 40*40;
begin
\Initial coordinates of centers of atoms
\Ethanol:  C      C       H       H       H       H       H       H       O
AtomX:= [-0.770, 0.770, -1.290, -1.290, -1.290,  1.290,  1.290,  2.250,  1.290];
AtomY:= [ 0.000, 0.000,  0.000, -0.778,  0.778, -0.778,  0.778,  0.000,  0.000];
AtomZ:= [ 0.000, 0.000,  1.040, -0.520, -0.520,  0.520,  0.520, -1.040, -1.040];
AtomR:= [ 5.000, 5.000,  2.500,  2.500,  2.500,  2.500,  2.500,  2.500,  3.750];
AtomC:= [ 1,	 1,	  3,	  3,	  3,	  3,	  3,	  3,	  2];
Bond:=  [2, 0,  3, 0,  4, 0,  0, 1,  1, 5,  1, 6,  1, 8,  7, 8];

\Set up color registers with gradients of color
port($3C8):= 64;		\starting color register 
	\	Red		Green		Blue
for I:= 0, 62 do
	[port($3C9):= I;   port($3C9):= I;   port($3C9):= 0];		\yellow
port($3C9):= I;   port($3C9):= I;   port($3C9):= I;			\hilite
for I:= 0, 62 do
	[port($3C9):= I;   port($3C9):= 0;   port($3C9):= 0];		\red
port($3C9):= I;   port($3C9):= I;   port($3C9):= I;			\hilite
for I:= 0, 62 do
	[port($3C9):= 0;   port($3C9):= I;   port($3C9):= I];		\cyan
port($3C9):= I;   port($3C9):= I;   port($3C9):= I;			\hilite

\Create atom (sphere) image
for J:= 0, 81-1 do
	begin
	Y:= J - 40;
	for I:= 0, 81-1 do
		begin
		X:= I - 40;
		C:= 0;				\assume background color
		Z2:= R2 - (X*X + Y*Y);
		if Z2 >= 0 then
			begin
			Z:= fix(sqrt(float(Z2)));
			C:= Z + 23;		\sphere's intensity (tweaked)
			end;
		Sphere(I+J*81):= C;
		end;
	end;
end;	\InitMole;

\################################# RECURSE #####################################

proc	Recurse;	\Now for something different: A picture of a picture...
int	X, Y;
begin
for Y:= 0, Edge-1 do		\sample every other dot of centered image
    for X:= 0, Edge-1 do
	Face(1, X+Y*Edge):= Image(X*2+(TVW>>1-Edge) + (Y*2+(TVH>>1-Edge))*TVW);
end;	\Recurse

\############################### STARFIELD #####################################

proc	GenStar(I);		\Sprinkle some stars into the firmament
int	I;
begin
StarX(I):= Ran(20000) - 10000;
StarY(I):= Ran(20000) - 10000;
StarZ(I):= Ran(10000) + 20000;	\far away random distance
end;	\GenStar



proc	DrawStars;		\Draw stars onto Image and move them
int	I, X, Y, C;


	proc	Dot(X, Y, C, D);	\Star size is a function of distance
	int	X, Y, C, D;

		proc	Point(X, Y, C);	\Redefine Point intrinsic
		int	X, Y, C;
		if X>=0 & X<TVW & Y>=0 & Y<TVH then	\clip to boundaries
			Image(X+Y*TVW):= C
		else GenStar(I);	\if star was clipped, generate a new one

	begin
	Point(X, Y, C);
	if D < 6000 then Point(X+1, Y+1, C);
	if D < 5000 then Point(X+1, Y, C);
	if D < 2500 then Point(X, Y+1, C);
	end;	\Dot


begin	\DrawStars
for I:= 0, StarN-1 do
	begin
	X:= StarX(I)<<10 / StarZ(I) + TVW/2;
	Y:= StarY(I)<<10 / StarZ(I) + TVH/2;
	C:= $1F - 16*StarZ(I)/30000;		\color depends on distance
	Dot(X, Y, C, StarZ(I));			\star size depends on distance
	end;

for I:= 0, StarN-1 do
	begin
	StarZ(I):= StarZ(I) - 100;		\move stars closer
	if StarZ(I) <= 100 then GenStar(I);	\if star passed by, gen new one
	end;
end;	\DrawStars



proc	InitStars;		\Initialize starfield
int	I;
begin
for I:= 0, StarN-1 do GenStar(I);
end;	\InitStars

\################################## CUBE #######################################

proc	Erase(Im);		\Erase Image buffer
int	Im;
int	I;
begin
for I:= -(TVW*TVH/IntSize-1), 0 do	\write four bytes at a time for speed
	Im(-I):= 0;			\going downward is slightly faster
end;	\Erase				 because of the memory cache



proc	DrawPoly(X0, Y0, VT, Tex, TVT);	\Draw a texture-mapped polygon on Image
\This procedure was originally written to draw any kind of polygon, convex or
\ concave. It was hastily hacked here to draw the parallelograms that form the
\ faces of the cube. The linked list could be replaced with a simple array
\ since there are always exactly two edges per scan line (unlike for a concave
\ polygon).
int	X0, Y0,		\position to draw first coordinate of polygon (pixels)
	VT;		\vertex table: 1-dimensional array of X,Y coordinates
			\ first value gives the number of vertexes (X,Y pairs)
char	Tex;		\array: texture image
int	TVT;		\texture vertex table, similar to VT

\    0  1  2  3  4  5  6  7  8
\VT: N  Px Py Qx Qy Rx Ry Sx Sy		array of 'polygon' vertexes
\TVT:-  px py qx qy rx ry sx sy		coorsponding array of texture vertexes

int	N, HeadTbl, P, MemPtr, \used to build linked list
	Gx, Gy, Hx, Hy,	\coordinates of end points of plotted horiz scan line
	I, J,		\indexes
	X, Y,		\coordinates to plot horiz scan lines
	YTVW		\Y * TVW (for speed)
	Min, Max,	\minimum and maximum Y coordinates of polygon
	Height;		\height of polygon in pixels
def	Nil= $80000000;	\linked list null pointer
def	SS = 16;	\scale shift factor
int	Py, Qy, Ry, Sy,	\Y coordinates of 4 corners of 'polygon'
	Tpx, Tpy, Tqx, Tqy, Trx, Try, Tsx, Tsy,	\texture map corner coordinates
	Cg, Ch,		\relative position of horizontal line from end points
	Tx0, Ty0, Tx1, Ty1, \end points of corresponding line in texture map
	Tdx, Tdy,	\amount to step coordinates in texture map
	Tx, Ty;		\coordinates in texture map


	proc	BuildLine(I0, I1);
	\Build table P containing X coordinates along polygon edge. This creates
	\ a linked list for each horizontal line that will be drawn through the
	\ polygon.
	int	I0, I1;	\indexes in VT specifying vertexes to build line between
	int	X, Y, DX, DY, M, T, Q;
	begin
	Gx:= VT(I0) + X0;   Gy:= VT(I0+1) + Y0;	\get coordinates of end points
	Hx:= VT(I1) + X0;   Hy:= VT(I1+1) + Y0;	\ of the polygon's edge

	if Gy > Hy then
		begin				\swap end points
		T:= Gy;   Gy:= Hy;   Hy:= T;
		T:= Gx;   Gx:= Hx;   Hx:= T;
		T:= I0;   I0:= I1;   I1:= T;
		end;

	DY:= Hy - Gy;
	if DY = 0 then return;			\don't draw a single horiz line
	DX:= Hx - Gx;
	M:= DX<<SS / DY;			\scaled slope (rotated 90 deg)

	X:= Gx<<SS;				\scaled starting X coordinate
	for Y:= Gy-Min, Hy-Min-1 do
		begin
		T:= X/(1<<SS);			\unscale X; (X may be negative)
		P:= HeadTbl + Y*IntSize;	\point to list header table
		loop	begin
			case of P(0)=Nil, P(0,1)>T: \if entry empty or greater
				begin		\insert item at front of list
				Q:= MemPtr;	\ (sorted left-to-right)
				MemPtr:= MemPtr + 3*IntSize;
				Q(0):= P(0);	\copy nil or pointer to larger X
				Q(1):= T;	\unscaled X
				Q(2):= I0 + I1<<8; \indexes of ends of line seg
				P(0):= Q;
				quit;
				end
			other	P:= P(0);	\follow pointers in linked list
			end;
		X:= X + M;			\next X coordinate
		end;
	end;	\BuildLine


begin	\DrawPoly
\Find maximum and minimum Y coordinates
N:= VT(0);				\get number of coordinates
Min:= 100000;   Max:= -100000;
I:= 2;					\index of first Y coordinate in VT
for J:= 1, N do
	begin
	Y:= Y0 + VT(I);
	if Y > Max then Max:= Y;
	if Y < Min then Min:= Y;
	I:= I + 2;			\next Y coordinate
	end;
Height:= Max - Min + 1;

\Set up array of list headers for each horizontal scan line in the polygon.
\ The list will contain the X coordinates (lowest value first) of the edges,
\ along with the index to the coordinates for the start and end of each edge.
HeadTbl:= Reserve(Height*IntSize);
for Y:= 0, Height-1 do HeadTbl(Y):= Nil;

MemPtr:= Reserve(400000);		\set up our own heap space manager
					\ (or in-line BuildLine)
I:= 1;					\index of first coordinate in array VT
for J:= 2, N do
	begin
	BuildLine(I, I+2);		\make edge of polygon between vertexes
	I:= I + 2;			\ i.e. build table of X values for edge
	end;
BuildLine(I, 1);			\make sure the polygon is closed, i.e.
					\ connect last vertex to first vertex
if Max > TVH-1 then Max:= TVH-1;	\clip to top and bottom of screen
for Y:= if Min<0 then 0 else Min, Max do\draw horizontal lines to fill polygon
	begin				\ from top to bottom
	P:= HeadTbl + (Y-Min)*IntSize;
	loop	begin			\for edge on left side of polygon...
		P:= P(0);
		if P = Nil then quit;
		Gx:= P(1);		\starting X coordinate for horiz line

		I:= P(2) & $00FF;	\index of start of edge
		Sy:= VT(I+1) + Y0;	\Y coordinate of start of edge
		Tsx:= TVT(I);		\corresponding coordinate in texture map
		Tsy:= TVT(I+1);

		I:= P(2)>>8;		\index of end of edge
		Ry:= VT(I+1) + Y0;	\Y coordinate of end of edge
		Trx:= TVT(I);		\corresponding coordinate in texture map
		Try:= TVT(I+1);

		P:= P(0);
		if P = Nil then quit;
		Hx:= P(1);		\ending X coordinate for horiz line

		if Gx # Hx then		\same as above for edge on right side
			begin		\ of polygon...
			I:= P(2) & $00FF;
			Py:= VT(I+1) + Y0;
			Tpx:= TVT(I);
			Tpy:= TVT(I+1);

			I:= P(2)>>8;
			Qy:= VT(I+1) + Y0;
			Tqx:= TVT(I);
			Tqy:= TVT(I+1);

			Cg:= (Y-Ry)<<SS / (Sy-Ry);	\relative position
			Ch:= (Y-Qy)<<SS / (Py-Qy);

			Tx0:= Cg * (Tsx-Trx) + Trx<<SS;	\corresponding end-
			Ty0:= Cg * (Tsy-Try) + Try<<SS;	\ points in texture map
			Tx1:= Ch * (Tpx-Tqx) + Tqx<<SS;
			Ty1:= Ch * (Tpy-Tqy) + Tqy<<SS;

			Tdx:= (Tx1-Tx0) / (Hx-Gx);	\amount to move in
			Tdy:= (Ty1-Ty0) / (Hx-Gx);	\ texture map

			Tx:= Tx0;	\coords of start of line in texture map
			Ty:= Ty0;
			YTVW:= Y*TVW;

			if Hx > TVW then Hx:= TVW;
			for X:= Gx, Hx-1 do	\step one horiz pixel at a time
				begin
				if X >= 0 then
				    Image(X + YTVW):= Tex(Tx>>SS + Edge*Ty>>SS);
				Tx:= Tx + Tdx; \loc of next pixel in texture map
				Ty:= Ty + Tdy;
				end;
			end;
		end;
	end;
end;	\DrawPoly

\-------------------------------------------------------------------------------

proc	DrawCube(X, Y, Size);	\Draw cube onto Image
int	X, Y;	\position of center of cube (pixels)
real	Size;	\magnification
int	T, TVT, CT, F, C, I;
begin
T:=   [4, 0,0, 0,0, 0,0, 0,0];		\table of 4 polygon vertex coordinates
TVT:= [0, 0,0, Edge-1,0, Edge-1,Edge-1, 0,Edge-1]; \CW from upper-right corner

\Table of corners: Indexes into CornerX, CornerY & CornerZ arrays for each face
CT:= [	[2, 0, 1, 3],	\0
	[4, 5, 1, 0],	\1
	[6, 4, 0, 2],	\2
	[7, 3, 1, 5],	\3
	[7, 6, 2, 3],	\4
	[5, 4, 6, 7] ];	\5

for F:= 0, 5 do				\for all the faces on the cube...
	begin
	I:= 1;
	for C:= 0, 3 do				\for all the corners on a face..
		begin
		T(I):= fix(CornerX(CT(F,C))*Size);	\X coordinate
		I:= I + 1;
		T(I):= fix(CornerY(CT(F,C))*Size);	\Y coordinate
		I:= I + 1;
		end;
	if CornerZ(CT(F,0)) + CornerZ(CT(F,2)) < 0.0 then
		\Z coordinate of center of polygon is visible so draw it
		DrawPoly(X, Y, T, addr Face(F,0), TVT);
	end;
end;	\DrawCube



proc	RotateCube;		\Rotate the cube
int	I;
def	Cx= 0.04/3., Cy= 0.05/3., Cz=-0.03/3.;		\rotation speed and axis
begin
for I:= 0, 7 do						\rotate all 8 corners...
	begin
	CornerY(I):= CornerY(I) - CornerZ(I)*Cx;	\rotate about X axis
	CornerZ(I):= CornerZ(I) + CornerY(I)*Cx;
	CornerZ(I):= CornerZ(I) - CornerX(I)*Cy;	\rotate about Y axis
	CornerX(I):= CornerX(I) + CornerZ(I)*Cy;
	CornerX(I):= CornerX(I) - CornerY(I)*Cz;	\rotate about Z axis
	CornerY(I):= CornerY(I) + CornerX(I)*Cz;
	end;
end;	\RotateCube

\-------------------------------------------------------------------------------

proc	MakeFaces;		\Make texture maps for the 6 die faces
int	F,	\face number
	P,	\pattern of spots
	Tbl,	\table of spot patterns for each face
	X, Y;	\coordinates


	proc	DrawCircle(X0, Y0);	\Draw a filled circle in Face array
	int	X0, Y0;		\coordinates of center (pixels)
	int	X, Y;		\working coordinates (pixels)
	def	R=28;		\radius (pixels)
	begin
	for Y:= -R, R do
	    for X:= -R, R do	\(speed is not an issue here)
		begin		\smooth the edge a little with gray
		if X*X+Y*Y < R*R then Face(F, X+X0 + Edge*(Y+Y0)):= $18;   \gray
		if X*X+Y*Y < R*R-25 then Face(F, X+X0 + Edge*(Y+Y0)):= $10;\black
		end;
	end;	\DrawCircle


begin
\Table to define location of spots on each face of a die. A spot is displayed at
\ the corresponding bit position:
\	0 1 2
\	3 4 5
\	6 7 8
Tbl:= [$010, $101, $111, $145, \$155\ $0E9, $1C7];

for F:= 0, 5 do				\for all the faces...
	begin
	for Y:= 0, Edge-1 do
	    for X:= 0, Edge-1 do	\shade of gray depends on face
		Face(F, X+Edge*Y):= abs(F+F-5)/2 + $1A;

	P:= Tbl(F);			\get pattern of spots for current face
	for Y:= 0, 2 do
	    for X:= 0, 2 do
		begin
		if P & 1 then DrawCircle(X*64+4+32, Y*64+4+32);
		P:= P >> 1;
		end;
	end;
end;	\MakeFaces

\-------------------------------------------------------------------------------

proc	ZoomCube;		\Rotate die and move it to the foreground
\Inputs: Bkgnd
int	X, Y, Z;
real	S;
begin
if ChkKey then return;
X:= 460;		\These constants have been carefully tweaked
Y:= 120;		\WARNING: Program can blow up if they are adjusted
Z:= 160;		\Protect your hard drive

FrameCount:= 0;
Time0:= GetTimeX;
repeat	begin
	Blit(DSeg, Bkgnd, DSeg, Image, TVW*TVH);

	S:= 0.6*160.0/float(Z);		\apparent size is function of distance
	DrawCube(X, Y, S);

	Wait26ms;
	WaitVB;
	Time0:= GetTimeX;
	Paint(0, 0, TVW, TVH, Image, TVW);	\copy image to screen
	if FrameCount = 0 then Sound(0, 27, 1);	\pause for effect

	X:= X - 1;
	Y:= Y + 1;
	Z:= Z - 1;
	RotateCube;
	FrameCount:= FrameCount + 1;
	end;
until FrameCount=158 ! ChkKey;
end;	\ZoomCube

\############################### BLUEPRINT #####################################

proc	Blueprint;		\Draw blueprint of a die (as in dice)
def	LegnX=TVW-39*8, LegnY= TVH-3*13-20;	\coordinates of legend (pixels)
def	DieS=160,		\length of side of die (pixels)
	Radius=20,		\radius of spot on die (pixels)
	TopX=100, TopY=40,	\coordinates of upper-left corner of views
	FrontX=TopX, FrontY=233,
	SideX=380, SideY=FrontY;
def	IsoX = SideX+DieS/2,	\coordinates of nearest point in isometric view
	IsoY = TopY+DieS/2,
	IsoH = 98,		\dimensions of isometric
	IsoW = IsoH*866/1000;	\cos(30 deg) = 0.866




proc	Erase(C);		\Erase screen
int	C;			\Color to fill screen with
int	X, Y;
begin
for Y:= 0, TVH-1 do
    for X:= 0, TVW-1 do
	Point(X, Y, C);
end;	\Erase



proc	LineTo(X, Y, C);	\Draw line from PenX,PenY to X,Y in color C
int	X, Y, C;	\coordinates of end-of-line, color
int	DX, DY,		\delta X, delta Y
	IX, IY,		\plotting increments along X and Y axes
	SY,		\flag: step along Y direction
	D,		\decision variable (move straight or diagonally)
	E, NE,		\East, Northeast (straight or diagonal)
	T, I;		\scratch
begin
DX:= X - PenX;				\delta X
if DX >= 0 then				\if positive then increment
	IX:= 1
else	[DX:= -DX;   IX:= -1];		\else absolute value and decrement

DY:= Y - PenY;				\same for Y direction
if DY >= 0 then
	IY:= 1
else	[DY:= -DY;   IY:= -1];

if DX >= DY then			\step along X direction
	SY:= false
else	begin
	SY:= true;			\step along Y direction
	T:= DX;   DX:= DY;   DY:= T;	\swap DX and DY
	end;

D:= 2*DY - DX;				\set up decision variable stuff
E:= 2*DY;
NE:= 2*(DY-DX);

Point(PenX, PenY, C);			\plot initial point
for I:= 1, DX do
	begin
	if D < 0 then
		begin			\move straight
		D:= D + E;
		if SY then PenY:= PenY + IY
		      else PenX:= PenX + IX;
		end
	else	begin			\move diagonally
		D:= D + NE;
		PenX:= PenX + IX;
		PenY:= PenY + IY;
		end;
	Delay(300);			\s-l-o-w-l-y
	Point(PenX, PenY, C);
	end;
end;	\LineTo



proc	Rectangle(X, Y, W, H, C); \Draw a rectangle
int	X, Y, W, H,	\upper left corner coordinates, width, height (pixels)
	C;		\color
begin
PenX:= X; PenY:= Y;
LineTo(X+W-1, Y, C);
LineTo(X+W-1, Y+H-1, C);
LineTo(X, Y+H-1, C);
LineTo(X, Y, C);
end;	\Rectangle



proc	Circle(X0, Y0, R, C);	\Draw a circle
int	X0, Y0,		\coordinates of center
	R,		\radius (pixels)
	C;		\line color
int	X, Y, E, U, V;

	proc	PlotOctants;
	begin
	WaitVB;			\draw s-l-o-w-l-y
	Point(X0+Y, Y0+X, C);
	Point(X0+X, Y0+Y, C);
	Point(X0-X, Y0+Y, C);
	Point(X0-Y, Y0+X, C);
	Point(X0-Y, Y0-X, C);
	Point(X0-X, Y0-Y, C);
	Point(X0+X, Y0-Y, C);
	Point(X0+Y, Y0-X, C);
	end;	\PlotOctants

begin	\Circle
X:= 0;
Y:= R;
U:= 1;
V:= 1 -R -R;
E:= 1 -R;
while X < Y do
	begin
	PlotOctants;
	if E < 0 then
		[U:= U +2;   V:= V +2;   E:= E +U]
	else	[U:= U +2;   V:= V +4;   E:= E +V;   Y:= Y -1];
	X:= X +1;
	end;
if X = Y then PlotOctants;
end;	\Circle



proc	CCircle(X0, Y0, R, C);	\Draw a circle with crosshairs
int	X0, Y0,		\coordinates of center
	R,		\radius (pixels)
	C;		\line color
begin
Circle(X0, Y0, R, C);

C:= C & 7;		\draw crosshairs in a darker shade
PenX:= X0-2*R;
PenY:= Y0;
LineTo(X0-R/2, Y0, C);
PenX:= X0+2*R;
LineTo(X0+R/2, Y0, C);
PenX:= X0-R/5;
LineTo(X0+R/5, Y0, C);

PenX:= X0;
PenY:= Y0-2*R;
LineTo(X0, Y0-R/2, C);
PenY:= Y0+2*R;
LineTo(X0, Y0+R/2, C);
PenY:= Y0-R/5;
LineTo(X0, Y0+R/5, C);
end;	\CCircle



proc	DrawChar(Ch, C);	\Display a character at PX,PY
int	Ch, C;			\upper-left coordinates, character
int	I, J, Byte;		\coordinates, size, indicies, font byte
begin
for J:= 0, FontHeight-1 do
	begin			\get byte from BIOS's font table
	Byte:= Peek(FontSeg,  FontOff + Ch*FontHeight + J);
	for I:= 0, 7 do		\for each bit in the Byte...
		begin
		if Byte & $80 then
			Point(PenX+I, PenY+J, C);
		Byte:= Byte << 1;
		end;
	end;
PenX:= PenX + 8;		\move to next character position
end;	\DrawChar



proc	DrawText(X, Y, Str, C);	\Draw all the characters in a text string
int	X, Y;	\coordinates (pixels)
char	Str;
int	C;
int	I, Ch;
begin
PenX:= X;   PenY:= Y;
for I:= 0, 1000 do
	begin
	Ch:= Str(I);
	if Ch >=$80 then
		[Ch:= Ch & $7F;   I:= 1000];
	WaitVB;			\draw s-l-o-w-l-y
	DrawChar(Ch, C);
	end;
end;	\DrawText



proc	DrawArrow(X, Y, C, D); \Draw an arrowhead
int	X, Y,	\coordinates at point of arrow head (pixels)
	C,	\color to draw
	D;	\direction arrowhead is pointing (0=right, 1=down, 2=left, 3=up)
int	I, J,	\indexes into Tbl
	B;	\byte from Tbl
char	Tbl;	\dot pattern for arrowhead
begin
Tbl:= [$FF0C3040, $0040300C];	\8x8 pattern of "->"
for J:= 0, 7 do
	begin
	B:= Tbl(J);
	for I:= 0, 7 do
		begin
		if B & $80>>I then
			case D of
			  0:	Point(X-7+I, Y-3+J, C);
			  1:	Point(X-3+J, Y-7+I, C);
			  2:	Point(X+7-I, Y-3+J, C);
			  3:	Point(X-3+J, Y+7-I, C)
			other	[];
		end;
	end;
end;	\DrawArrow



proc	DrawDimension(X0, Y0, X1, Y1, Str);	\Draw a dimension line
int	X0, Y0, X1, Y1,	\coordinates of end points (pixels) (ascending order)
	Str;		\5-char string that labels the distance
int	DX, DY;
begin
PenX:= X0;
PenY:= Y0;
DX:= (X1-X0)/2;
DY:= (Y1-Y0)/2;
if DY = 0 then
	begin
	LineTo(X0+DX-28, Y0, Cyan);
	PenX:= X0+DX+28;
	LineTo(X1, Y1, Cyan);
	PenX:= X0; PenY:= Y0-16;
	LineTo(X0, Y0+16, Cyan);
	PenX:= X1; PenY:= Y1-16;
	LineTo(X1, Y1+16, Cyan);
	DrawText(X0+DX-20, Y0-4, Str, Cyan);
	DrawArrow(X0+1, Y0, Cyan, 2);
	DrawArrow(X1-1, Y1, Cyan, 0);
	end
else	begin		\DX = 0
	LineTo(X0, Y0+DY-9, Cyan);
	PenY:= Y0+DY+8;
	LineTo(X1, Y1, Cyan);
	PenX:= X0-16; PenY:= Y0;
	LineTo(X0+16, Y0, Cyan);
	PenX:= X1-16; PenY:= Y1;
	LineTo(X1+16, Y1, Cyan);
	DrawText(X0-20, Y0+DY-3, Str, Cyan);
	DrawArrow(X0, Y0+1, Cyan, 3);
	DrawArrow(X1, Y1-1, Cyan, 1);
	end;
end;	\DrawDimension



proc	Flood(X, Y, C, C0);	\Fill an area of color C0 with color C
int	X, Y,	\seed coordinate (where to start)
	C, C0;	\color to fill with and color to replace
def	S=8000;	\size of queue (must be an even number)
int	F, E,	\fill and empty indexes
	Q(S);	\queue (FIFO)


	proc	EnQ(X, Y);	\Enqueue coordinate
	int	X, Y;
	begin
	Q(F):= X;
	F:= F + 1;
	Q(F):= Y;
	F:= F + 1;
	if F >= S then F:= 0;
	end;	\EnQ


begin	\Flood
F:= 0;   E:= 0;
EnQ(X, Y);
while E # F do
	begin
	X:= Q(E);		\dequeue coordinate
	E:= E + 1;
	Y:= Q(E);
	E:= E + 1;
	if E >= S then E:= 0;

	if ReadPix(X, Y) = C0 then
		begin
		Point(X, Y, C);
		Delay(70);	\s-l-o-w-l-y
		EnQ(X+1, Y);	\enqueue adjacent pixels
		EnQ(X-1, Y);
		EnQ(X, Y+1);
		EnQ(X, Y-1);
		end;
	end;
end;	\Flood



begin	\Blueprint
Erase(Blue);
Sound(0, 18, 1);	\delay about 1 second
if ChkKey then return;

\Draw border
Rectangle(4, 4, TVW-8, TVH-8, LCyan);

\Draw legend lines
Rectangle(LegnX, LegnY-4, TVW-6-LegnX, TVH-6-LegnY+4, LCyan);
PenX:= LegnX+1;   PenY:= LegnY+17-3;
LineTo(TVW-6-1, LegnY+17-3, LCyan);
PenX:= LegnX+1;   PenY:= LegnY+17+13-3;
LineTo(TVW-6-2, LegnY+17+13-3, Cyan);
PenX:= LegnX+1;   PenY:= LegnY+17+26-3;
LineTo(TVW-6-2, LegnY+17+26-3, Cyan);

PenX:= LegnX+18*8;   PenY:= LegnY+17-2;
LineTo(LegnX+18*8, LegnY+17+13-3, Cyan);
PenX:= LegnX+28*8;   PenY:= LegnY+17+13-2;
LineTo(LegnX+28*8, TVH-8, Cyan);
if ChkKey then return;

Rectangle(FrontX, FrontY, DieS, DieS, LCyan);
Rectangle(SideX, SideY, DieS, DieS, LCyan);
Rectangle(TopX, TopY, DieS, DieS, LCyan);
if ChkKey then return;

CCircle(TopX+DieS/6, TopY+DieS/6, Radius, LCyan);
CCircle(TopX+DieS*5/6, TopY+DieS*5/6, Radius, LCyan);
CCircle(TopX+DieS*5/6, TopY+DieS/6, Radius, LCyan);
CCircle(TopX+DieS/6, TopY+DieS*5/6, Radius, LCyan);
CCircle(TopX+DieS/2, TopY+DieS/6, Radius, LCyan);
CCircle(TopX+DieS/2, TopY+DieS*5/6, Radius, LCyan);
if ChkKey then return;

CCircle(FrontX+DieS/6, FrontY+DieS/6, Radius, LCyan);
CCircle(FrontX+DieS*5/6, FrontY+DieS*5/6, Radius, LCyan);

CCircle(SideX+DieS/6, SideY+DieS/6, Radius, LCyan);
CCircle(SideX+DieS*5/6, SideY+DieS*5/6, Radius, LCyan);
CCircle(SideX+DieS*5/6, SideY+DieS/6, Radius, LCyan);
CCircle(SideX+DieS/6, SideY+DieS*5/6, Radius, LCyan);
if ChkKey then return;

DrawDimension(FrontX, FrontY+DieS+30, FrontX+DieS-1, FrontY+DieS+30, "160.0");
DrawDimension(FrontX+DieS+30, FrontY, FrontX+DieS+30, FrontY+DieS-1, "160.0");
DrawDimension(SideX-30, SideY+5*DieS/6, SideX-30, SideY+DieS-1, "26.67");

\Draw legend text
DrawText(LegnX, LegnY+2,     "       Cubic, a 32-bit XPL0 Demo", Cyan);
DrawText(LegnX, LegnY+17,    "  Revision: 1.0     Date: 17-Jan-2005", Cyan);
if ChkKey then return;
DrawText(LegnX, LegnY+17+13, "  Drawn by: Lenny Boreal     Size:  A", Cyan);
DrawText(LegnX, LegnY+17+26, "  loren_blaney@idcomm.com    Sheet: 1", Cyan);
if ChkKey then return;

\Save image for background
for PenY:= 0, TVH-1 do
    for PenX:= 0, TVW-1 do
	Bkgnd(PenX+PenY*TVW):= ReadPix(PenX, PenY);

\Draw isometric				corner
PenX:= IsoX;	PenY:= IsoY;		\3
LineTo(IsoX+IsoW, IsoY-IsoH/2, Cyan);	\2
LineTo(IsoX, IsoY-IsoH, Cyan);		\6
LineTo(IsoX-IsoW, IsoY-IsoH/2, Cyan);	\7
LineTo(IsoX, IsoY, Cyan);		\3
LineTo(IsoX, IsoY+IsoH, Cyan);		\1
LineTo(IsoX-IsoW, IsoY+IsoH/2, Cyan);	\5
LineTo(IsoX-IsoW, IsoY-IsoH/2, Cyan);	\7
PenX:= IsoX;	PenY:= IsoY+IsoH;	\1
LineTo(IsoX+IsoW, IsoY+IsoH/2, Cyan);	\0
LineTo(IsoX+IsoW, IsoY-IsoH/2, Cyan);	\2
if ChkKey then return;

Flood(IsoX-1, IsoY+IsoH-2, abs(1+1-5)/2+$1A, Blue);	\2 spots
if ChkKey then return;
Flood(IsoX+1, IsoY+IsoH-2, abs(3+3-5)/2+$1A, Blue);	\4 spots
if ChkKey then return;
Flood(IsoX,   IsoY-1,      abs(5+5-5)/2+$1A, Blue);	\6 spots
end;	\Blueprint

\################################### MAIN ######################################
\
\			    6 __________ 2
\			    /|         /|		    Z
\			  /_________ /  |		  /              
\			7|   |      |3  |	        +----- X
\			 |          |   |		|
\			 |  4|_ _ _ | _ |		|
\			 |  /       |  / 0		Y
\			 |__________|/ 
\			5            1
\
begin	\Main
CpuReg:= GetReg;
DSeg:= CpuReg(12);

FreeMem:= Free;

\Coordinates defining the isometric positions of the 8 corners of a cube:
\		   0	   1	   2	   3	   4	   5	   6	   7
CornerX:=      [-141.4,   -0.0,    0.0,  141.4, -141.4,   -0.0,    0.0,  141.4];
CornerY:=      [  81.6,  163.3,   -0.0,   81.6,  -81.6,    0.0, -163.3,  -81.6];
CornerZ:=      [  57.7,  -57.7,  173.2,   57.7,  -57.7, -173.2,   57.7,  -57.7];

OpenTimer;
OpenPM;
TrapC(true);		\make sure Ctrl+C does not bypass a proper exit

SetVid(TVMode);		\set graphic mode
InitListing;		\(do this now in case input file is not found)
InitScroller;		\get font for Scroller, also Listing and Blueprint
Blueprint;
MakeFaces;
ZoomCube;

InitLife;
InitNoise;
InitStars;
InitPlot3D;
InitMole;

\Trap(false);		turn off error trapping
\Coordinates of cube in its standard orientation:
\		   0	   1	   2	   3	   4	   5	   6	   7
CornerX:=      [ 100.0,	 100.0,	 100.0,	 100.0,	-100.0,	-100.0,	-100.0,	-100.0];
CornerY:=      [ 100.0,	 100.0,	-100.0,	-100.0,	 100.0,	 100.0,	-100.0,	-100.0];
CornerZ:=      [ 100.0,	-100.0,	 100.0,	-100.0,	 100.0,	-100.0,	 100.0,	-100.0];

\Measure the frame rate (the time between vertical blank signals):
WaitVB;
StartTime:= GetTimeX;		\start timer for measuring loop time
WaitVB;
FrameTime:= ElapsedTime;

\Fly through some stars for awhile:
FrameCount:= 0;
Time0:= GetTimeX;
repeat	begin
	Erase(Image);
	DrawStars;
	Wait26ms;
	WaitVB;
	Time0:= GetTimeX;
	Paint(0, 0, TVW, TVH, Image, TVW);		\copy image to screen
	FrameCount:= FrameCount + 1;
	end;
until FrameCount=120 ! ChkKey;

\Move cube to foreground then startup the various animations on each face
FrameCount:= 0;
CubeX:= 0;	CubeY:= 0;	CubeZ:= 15000;
CubeVx:= 1.0;	CubeVy:= 0.0;	CubeVz:= 0.0;
Time0:= GetTimeX;
StartTime:= GetTimeX;	\initialize timer for measuring loop time
repeat	begin
	if FrameCount>=245 & FrameCount<260 then Noise(3);
	if FrameCount = 260 then 
		begin
		for CurY:= 0, Edge-1 do			\blank the cube face
		    for CurX:= 0, Edge-1 do
			Face(3, CurX+CurY*Edge):= $12;	\dark gray
		CurX:= 0;		\start text cursor on last line
		CurY:= (Edge/8 - 1)*8;
		end;
	if FrameCount > 260 then Listing;

	if FrameCount>=365 & FrameCount<380 then Noise(2);
	if FrameCount > 380 then Plot3D;

	if FrameCount = 480 then InitKaleid;		\(erases face)
	if FrameCount > 480 then Kaleid;

	if FrameCount > 580 then 
		if rem(FrameCount/3) = 0 then Scroller;

	if FrameCount > 780 then Molecule;

	if FrameCount > 980 then Recurse
	else if rem(FrameCount/3) = 1 then CycleLife;

	Erase(Image);
	DrawStars;

	\Cube's velocity is a function of its distance from the center
	CubeVx:= CubeVx + float(TVW/2-CubeX)/500.0;
	CubeVy:= CubeVy + float(TVH/2-CubeY)/600.0;
	CubeVz:= CubeVz + float(250-CubeZ)/1000.0;	\arbitrary distance

	if FrameCount < 240 then	\apply dampening else let it bobble
		begin
		CubeVx:= CubeVx * 0.99;
		CubeVy:= CubeVy * 0.98;	\WARNING:
		CubeVz:= CubeVz * 0.95;	\ program can blow if values are tweaked
		end;			\ (protect your hard drive)

	\Cube's position is a function of its velocity
	CubeX:= CubeX + fix(CubeVx);
	CubeY:= CubeY + fix(CubeVy);
	CubeZ:= CubeZ + fix(CubeVz);

	CubeSize:= 400.0/float(CubeZ);	\apparent size is function of distance
	DrawCube(CubeX, CubeY, CubeSize);

	if ShiftKey then DrawPM(ElapsedTime);

	Wait26ms;
	WaitVB;
	Time0:= GetTimeX;		\start timer for 26 ms delay
	StartTime:= GetTimeX;		\start timer for measuring loop time
	Paint(0, 0, TVW, TVH, Image, TVW);		\copy Image to screen

	RotateCube;
	FrameCount:= FrameCount + 1;
	end;
until ChkKey;					\keystroke terminates program
OpenI(0);					\eat the keystroke
Fadeout;
SetVid($03);					\restore text display and colors
CloseTimer;
end;	\Main
